# -*- coding: UTF-8 -*-
import unittest
from mock import patch, Mock
import json
from pandasticsearch.dataframe import DataFrame, Column
from pandasticsearch.operators import *


@patch('pandasticsearch.client.urllib.request.urlopen')
def create_df_from_es(mock_urlopen):
    response = Mock()
    dic = {
        "index": {
            "mappings": {
                "doc_type": {
                    "properties": {
                        "a": {
                            "type": "integer"
                        },
                        "b": {
                            "type": "integer"
                        },
                        "c": {"properties": {
                            "d": {"type": "keyword", "ignore_above": 1024},
                            "e": {"type": "keyword", "ignore_above": 1024}
                        }
                        }
                    }
                }
            }
        }
    }
    response.read.return_value = json.dumps(dic).encode("utf-8")
    mock_urlopen.return_value = response
    return DataFrame.from_es(url="http://localhost:9200", index='index', doc_type='doc_type')


@patch('pandasticsearch.client.urllib.request.urlopen')
def create_df_from_es_after_removal_mapping(mock_urlopen):
    """
    https://www.elastic.co/guide/en/elasticsearch/reference/current/removal-of-types.html
    """
    response = Mock()
    dic = {
        "index": {
            "mappings": {
                "properties": {
                    "a": {
                        "type": "integer"
                    },
                    "b": {
                        "type": "integer"
                    },
                    "c": {"properties": {
                        "d": {"type": "keyword", "ignore_above": 1024},
                        "e": {"type": "keyword", "ignore_above": 1024}
                    }
                    }
                }
            }
        }
    }
    response.read.return_value = json.dumps(dic).encode("utf-8")
    mock_urlopen.return_value = response
    return DataFrame.from_es(url="http://localhost:9200", index='index', compat=7)


class TestDataFrame(unittest.TestCase):
    def test_getitem(self):
        df = create_df_from_es()
        self.assertTrue(isinstance(df['a'], Column))
        self.assertTrue(isinstance(df['b'], Column))

        expr = df['a'] > 2
        self.assertTrue(isinstance(expr, BooleanFilter))
        self.assertTrue(isinstance(df[expr], DataFrame))
        self.assertEqual(df[expr]._filter.build(), {'range': {'a': {'gt': 2}}})

    def test_getattr(self):
        df = create_df_from_es()
        self.assertTrue(isinstance(df.a, Column))
        self.assertTrue(isinstance(df.b, Column))

    def test_index(self):
        df = create_df_from_es()
        self.assertEqual(df.index, 'index/doc_type')

    def test_index_after_removal_mapping(self):
        df = create_df_from_es_after_removal_mapping()
        self.assertEqual(df.index, 'index')

    def test_columns(self):
        df = create_df_from_es()
        self.assertEqual(df.columns, ['a', 'b', 'c.d', 'c.e'])

    def test_print_schema(self):
        df = create_df_from_es()
        df.print_schema()

    def test_columns_after_removal_mapping(self):
        df = create_df_from_es_after_removal_mapping()
        self.assertEqual(df.columns, ['a', 'b', 'c.d', 'c.e'])

    def test_print_schema_after_removal_mapping(self):
        df = create_df_from_es_after_removal_mapping()
        df.print_schema()

    def test_init(self):
        df = create_df_from_es()
        self.assertEqual(df.to_dict(), {'size': 20})

    def test_filter(self):
        df = create_df_from_es()

        self.assertEqual((df.filter(df['a'] > 2)).to_dict(),
                         {'query': {'filtered': {'filter': {'range': {'a': {'gt': 2}}}}},
                          'size': 20})

        self.assertEqual((df.filter((df['a'] > 2) & (df.b == 1))).to_dict(),
                         {'query': {'filtered': {'filter': {'bool': {'must': [
                             {'range': {'a': {'gt': 2}}},
                             {'term': {'b': 1}}]}}}},
                             'size': 20})

        self.assertEqual((df.filter(df['a'] > 2).filter(df.b == 1)).to_dict(),
                         {'query': {'filtered': {'filter': {'bool': {'must': [
                             {'range': {'a': {'gt': 2}}},
                             {'term': {'b': 1}}]}}}},
                             'size': 20})

        self.assertEqual(df.where(Greater('a', 2)).to_dict(),
                         {'query': {'filtered': {'filter': {'range': {'a': {'gt': 2}}}}},
                          'size': 20})

        self.assertEqual(df.filter('2016 - doc["age"].value > 1995').to_dict(),
                         {'query': {'filtered': {
                             'filter': {'script': {'script': {'inline': '2016 - doc["age"].value > 1995'}}}}},
                             'size': 20})

    def test_groupby(self):
        df = create_df_from_es()
        self.assertEqual((df.groupby(df.a)).to_dict(),
                         {'aggregations': {'a': {'terms': {'field': 'a', 'size': 20}}}, 'size': 0})

        self.assertEqual((df.groupby(df['a'], df['b'])).to_dict(),
                         {
                             'aggregations': {
                                 'a': {
                                     'aggregations': {
                                         'b': {
                                             'terms': {'field': 'b', 'size': 20}}
                                     },
                                     'terms': {'field': 'a', 'size': 20}}},
                             'size': 0})

    def test_agg(self):
        df = create_df_from_es()
        self.assertEqual((df.agg(MetricAggregator('a', 'avg'))).to_dict(),
                         {'aggregations': {'avg(a)': {'avg': {'field': 'a'}}}, 'size': 0})

    def test_sort(self):
        df = create_df_from_es()
        self.assertEqual((df.sort(df['a'].asc)).to_dict(),
                         {'sort': [{'a': {'order': 'asc'}}], 'size': 20})

        self.assertEqual((df.sort(Sorter('a'), Sorter('b'))).to_dict(),
                         {'sort': [{'a': {'order': 'desc'}},
                                   {'b': {'order': 'desc'}}], 'size': 20})

        self.assertEqual((df.sort('doc["age"].value * 2')).to_dict(),
                         {'sort': [{'_script': {
                             'order': 'desc',
                             'script': 'doc["age"].value * 2',
                             'type': 'number'
                         }}], 'size': 20})

    def test_select(self):
        df = create_df_from_es()
        self.assertEqual(df.select('a').to_dict(),
                         {'_source': {'excludes': [], 'includes': ['a']}, 'size': 20})

        self.assertEqual(df.select(df['a'], df['b']).to_dict(),
                         {'_source': {'excludes': [], 'includes': ['a', 'b']}, 'size': 20})

    def test_limit(self):
        df = create_df_from_es()
        self.assertEqual(df.limit(199).to_dict(), {'size': 199})

    def test_complex(self):
        df = create_df_from_es()

        df2 = df.filter(df['a'] > 2)
        df3 = df2.select('a').limit(30)

        print(df3.to_dict())
        self.assertEqual(df3.to_dict(),
                         {'_source': {'excludes': [], 'includes': ['a']},
                          'query': {'filtered': {'filter': {'range': {'a': {'gt': 2}}}}},
                          'size': 30})

        df4 = df3.groupby('b')
        df5 = df4.agg(MetricAggregator('a', 'avg'))

        print(df5.to_dict())

        self.assertEqual(df5.to_dict(),
                         {'_source': {'excludes': [], 'includes': ['a']},
                          'aggregations': {
                              'b': {
                                  'terms': {'field': 'b', 'size': 20},
                                  'aggregations': {
                                      'avg(a)': {'avg': {'field': 'a'}}}}
                          },
                          'query': {'filtered': {'filter': {'range': {'a': {'gt': 2}}}}},
                          'size': 0})

        df6 = df5.sort(Sorter('a'))

        print(df6.to_dict())
        self.assertEqual(df6.to_dict(),
                         {'_source': {'excludes': [], 'includes': ['a']},
                          'aggregations': {
                              'b': {
                                  'terms': {'field': 'b', 'size': 20},
                                  'aggregations': {
                                      'avg(a)': {'avg': {'field': 'a'}}}}
                          },
                          'query': {'filtered': {'filter': {'range': {'a': {'gt': 2}}}}},
                          'sort': [{'a': {'order': 'desc'}}],
                          'size': 0})

    def test_complex_agg(self):
        df = create_df_from_es()
        df2 = df.groupby(df.b, df.a)

        df3 = df2.agg(MetricAggregator('a', 'avg'))

        self.assertEqual(df3.to_dict(),
                         {
                             'size': 0,
                             'aggregations': {
                                 'b': {
                                     'terms': {'field': 'b', 'size': 20},
                                     'aggregations': {
                                         'a': {
                                             'terms': {'field': 'a', 'size': 20},
                                             'aggregations': {
                                                 'avg(a)': {'avg': {'field': 'a'}}}}
                                     }}}})


if __name__ == '__main__':
    unittest.main()



# -*- coding: UTF-8 -*-
import unittest
from mock import patch, Mock

from pandasticsearch.client import RestClient


class TestClients(unittest.TestCase):
    @patch('pandasticsearch.client.urllib.request.urlopen')
    def test_rest_client_returns_results(self, mock_urlopen):
        response = Mock()
        response.read.return_value = """{"hits" : {"hits": [{"_source": {}}] }}""".encode("utf-8")
        mock_urlopen.return_value = response

        client = RestClient("http://localhost:9200")

        json = client.post(data="xxxx")

        print(json)
        self.assertIsNotNone(json)
        self.assertEqual(json, {"hits": {"hits": [{"_source": {}}]}})


if __name__ == '__main__':
    unittest.main()



# -*- coding: UTF-8 -*-
import unittest

from pandasticsearch.queries import Select, Agg


def create_hits():
    return {
        'hits': {
            'hits': [
                {'_source': {'a': 1, 'b': 1}},
                {'_source': {'a': 2, 'b': 2}},
                {'_source': {'a': 3, 'b': 3}},
            ]
        },
        'took': 1
    }


class TestQueries(unittest.TestCase):
    def test_select_explain_result(self):
        select = Select()
        select._result_dict = create_hits()
        select.explain_result()
        print(select)
        print(repr(select))

        self.assertIsNotNone(select.result)
        self.assertEqual(len(select), 3)

    def test_select_from_dict(self):
        select = Select.from_dict(create_hits())
        print(select)
        print(repr(select))

        self.assertIsNotNone(select.result)
        self.assertEqual(len(select), 3)

    def test_select_result(self):
        select = Select.from_dict(create_hits())
        print(select.result)

        self.assertIsNotNone(select.result)
        self.assertEqual(len(select.result[0]), 2)

    def test_agg_buckets(self):
        agg = Agg()
        agg._result_dict = {
            'took': 1,
            'aggregations': {
                'agg_key': {
                    'buckets': [
                        {
                            'key': 'a',
                            'f1': {'value': 100},
                            'f2': {'value': 1},
                            "doc_count": 12
                        },
                        {
                            'key': 'b',
                            'f1': {'value': 200},
                            'f2': {'value': 2},
                            "doc_count": 13
                        },
                    ]
                }
            }
        }

        agg.explain_result()
        print(agg.result)
        self.assertEqual(agg.result, [{'f1': 100, 'f2': 1, 'doc_count': 12},
                                      {'f1': 200, 'f2': 2, 'doc_count': 13}])
        self.assertEqual(agg.index, [('a',),
                                     ('b',)])

    def test_agg_date_histogram(self):
        agg = Agg()
        agg._result_dict = {
            'took': 3,
            'aggregations': {
                'my_date_histogram': {
                    'buckets': [
                        {
                            'doc_count': 1,
                            'key': 1480392360000,
                            'key_as_string': '2016-11-29T04:06:00.000Z',
                            'f1': {'value': 0.1},
                        },
                        {
                            'doc_count': 1,
                            'key': 1480392420000,
                            'key_as_string': '2016-11-29T04:07:00.000Z',
                            'f1': {'value': 0.2},
                        },
                        {
                            'doc_count': 1,
                            'key': 1480392480000,
                            'key_as_string': '2016-11-29T04:08:00.000Z',
                            'f1': {'value': 0.3},
                        },
                        {
                            'doc_count': 1,
                            'key': 1480392540000,
                            'key_as_string': '2016-11-29T04:09:00.000Z',
                            'f1': {'value': 0.4},
                        },
                        {
                            'doc_count': 1,
                            'key': 1480392600000,
                            'key_as_string': '2016-11-29T04:10:00.000Z',
                            'f1': {'value': 0.5},
                        }
                    ]
                }
            }
        }
        agg.explain_result()
        print(agg.result)
        self.assertEqual(agg.result, [{'f1': 0.1, 'doc_count': 1},
                                      {'f1': 0.2, 'doc_count': 1},
                                      {'f1': 0.3, 'doc_count': 1},
                                      {'f1': 0.4, 'doc_count': 1},
                                      {'f1': 0.5, 'doc_count': 1}])
        self.assertEqual(agg.index, [('2016-11-29T04:06:00.000Z',),
                                     ('2016-11-29T04:07:00.000Z',),
                                     ('2016-11-29T04:08:00.000Z',),
                                     ('2016-11-29T04:09:00.000Z',),
                                     ('2016-11-29T04:10:00.000Z',)])

    def test_agg_nested_buckets(self):
        agg = Agg()
        agg._result_dict = {
            'took': 1,
            'aggregations': {
                'agg_key1': {
                    'buckets': [
                        {
                            'key': 'a',
                            'agg_key2': {
                                'buckets': [
                                    {
                                        'key': 'x',
                                        'f1': {'value': 100},
                                        'f2': {'value': 1},
                                        "doc_count": 11
                                    },
                                    {
                                        'key': 'y',
                                        'f1': {'value': 200},
                                        'f2': {'value': 2},
                                        "doc_count": 12
                                    },
                                ]
                            }
                        },
                        {
                            'key': 'b',
                            'agg_key2': {
                                'buckets': [
                                    {
                                        'key': 'x',
                                        'f1': {'value': 300},
                                        'f2': {'value': 3},
                                        "doc_count": 13
                                    },
                                    {
                                        'key': 'y',
                                        'f1': {'value': 400},
                                        'f2': {'value': 4},
                                        "doc_count": 14
                                    },
                                ]
                            }
                        }
                    ]
                }
            }
        }

        agg.explain_result()
        print(agg.result)

        self.assertEqual(agg.result,
                         [{'f1': 100, 'f2': 1, 'doc_count': 11},
                          {'f1': 200, 'f2': 2, 'doc_count': 12},
                          {'f1': 300, 'f2': 3, 'doc_count': 13},
                          {'f1': 400, 'f2': 4, 'doc_count': 14}])
        self.assertEqual(agg.index, [('a', 'x'),
                                     ('a', 'y'),
                                     ('b', 'x'),
                                     ('b', 'y')])


if __name__ == '__main__':
    unittest.main()



# -*- coding: UTF-8 -*-


class PandasticSearchException(RuntimeError):
    def __init__(self, msg):
        super(PandasticSearchException, self).__init__(msg)


class NoSuchDependencyException(PandasticSearchException):
    pass


class ServerDefinedException(PandasticSearchException):
    pass


class ParseResultException(PandasticSearchException):
    pass


class DataFrameException(PandasticSearchException):
    pass



import sys

info = sys.version_info
import unittest2


try:
    callable = callable
except NameError:
    def callable(obj):
        return hasattr(obj, '__call__')


with_available = sys.version_info[:2] >= (2, 5)


def is_instance(obj, klass):
    """Version of is_instance that doesn't access __class__"""
    return issubclass(type(obj), klass)


class SomeClass(object):
    class_attribute = None

    def wibble(self):
        pass


class X(object):
    pass

try:
    next = next
except NameError:
    def next(obj):
        return obj.next()



# -*- coding: UTF-8 -*-

import json
import sys
import base64
import ssl
from six.moves import urllib

from pandasticsearch.errors import ServerDefinedException


class RestClient(object):
    """
    RestClient talks to Elasticsearch cluster through native RESTful API.
    """

    def __init__(self, url, endpoint='', username=None, password=None, verify_ssl=True):
        """
        Initialize the RESTful from the keyword arguments.

        :param str url: URL of Broker node in the Elasticsearch cluster
        :param str endpoint: Endpoint that Broker listens for queries on
        """
        self.url = url
        self.endpoint = endpoint
        self.username = username
        self.password = password
        self.verify_ssl = verify_ssl

    def _prepare_url(self):
        if self.url.endswith('/'):
            url = self.url + self.endpoint
        else:
            url = self.url + '/' + self.endpoint
        return url

    def get(self, params=None):
        """
        Sends a GET request to Elasticsearch.

        :param optional params: Dictionary to be sent in the query string.
        :return: The response as a dictionary.

        >>> from pandasticsearch import RestClient
        >>> client = RestClient('http://localhost:9200', '_mapping/index')
        >>> print(client.get())
        """
        try:
            url = self._prepare_url()
            username = self.username
            password = self.password
            verify_ssl = self.verify_ssl

            if params is not None:
                url = '{0}?{1}'.format(url, urllib.parse.urlencode(params))

            req = urllib.request.Request(url=url)

            if username is not None and password is not None:
                base64creds = base64.b64encode('%s:%s' % (username,password))
                req.add_header("Authorization", "Basic %s" % base64creds)
            
            if verify_ssl is False:
                context = ssl._create_unverified_context()
                res = urllib.request.urlopen(req, context=context)
            else:
                res = urllib.request.urlopen(req)

            data = res.read().decode("utf-8")
            res.close()
        except urllib.error.HTTPError:
            _, e, _ = sys.exc_info()
            reason = None
            if e.code != 200:
                try:
                    reason = json.loads(e.read().decode("utf-8"))
                except (ValueError, AttributeError, KeyError):
                    pass
                else:
                    reason = reason.get('error', None)

            raise ServerDefinedException(reason)
        else:
            return json.loads(data)

    def post(self, data, params=None):
        """
        Sends a POST request to Elasticsearch.

        :param data: The json data to send in the body of the request.
        :param optional params: Dictionary to be sent in the query string.
        :return: The response as a dictionary.

        >>> from pandasticsearch import RestClient
        >>> client = RestClient('http://localhost:9200', 'index/type/_search')
        >>> print(client.post(data={"query":{"match_all":{}}}))
        """
        try:
            url = self._prepare_url()
            username = self.username
            password = self.password
            verify_ssl = self.verify_ssl

            if params is not None:
                url = '{0}?{1}'.format(url, urllib.parse.urlencode(params))

            req = urllib.request.Request(url=url, data=json.dumps(data).encode('utf-8'),
                                         headers={'Content-Type': 'application/json'})

            if username is not None and password is not None:
                base64creds = base64.b64encode('%s:%s' % (username,password))
                req.add_header("Authorization", "Basic %s" % base64creds)
            
            if verify_ssl is False:
                context = ssl._create_unverified_context()
                res = urllib.request.urlopen(req, context=context)
            else:
                res = urllib.request.urlopen(req)

            data = res.read().decode("utf-8")
            res.close()
        except urllib.error.HTTPError:
            _, e, _ = sys.exc_info()
            reason = None
            if e.code != 200:
                try:
                    reason = json.loads(e.read().decode("utf-8"))
                except (ValueError, AttributeError, KeyError):
                    pass
                else:
                    reason = reason.get('error', None)

            raise ServerDefinedException(reason)
        else:
            return json.loads(data)



# -*- coding: UTF-8 -*-


class Grouper(object):
    def __init__(self, field, size=20, inner=None, include=None, exclude=None):
        self._field = field
        self._size = size
        self._inner = inner
        self._include = include
        self._exclude = exclude

    @staticmethod
    def from_list(l):
        if len(l) == 1:
            return Grouper(l[0])
        return Grouper(l[0], inner=Grouper.from_list(l[1:]))

    def build(self):
        terms = {'field': self._field, 'size': self._size}
        if self._exclude is not None:
            assert isinstance(self._exclude, list)
            terms['exclude'] = self._exclude
        if self._include is not None:
            assert isinstance(self._include, list)
            terms['include'] = self._include

        agg = {"terms": terms}

        if self._inner is not None:
            agg["aggregations"] = self._inner.build()

        return {self._field: agg}


class RangeGrouper(Grouper):
    def __init__(self, field, range_list):
        assert isinstance(range_list, list)
        super(RangeGrouper, self).__init__(field)
        self._field = field
        self._range_list = range_list

    def build(self):
        ranges = []
        starts = self._range_list[:-1]
        ends = self._range_list[1:]
        for start, end in zip(starts, ends):
            ranges.append({'from': start, 'to': end})

        name = 'range(' + ','.join([str(x) for x in self._range_list]) + ')'
        return {name: {'range': {'field': self._field, 'ranges': ranges}}}


class DateGrouper(Grouper):
    def __init__(self, field, interval, format):
        super(DateGrouper, self).__init__(field)
        self._field = field
        self._interval = interval
        self._format = format

    def build(self):
        name = 'date({0},{1})'.format(self._field, self._interval)
        return {name: {'date_histogram': {
            'field': self._field,
            'interval': self._interval,
            'format': self._format,
        }}}



# Copyright (c) 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright (C) 2013 Association of Universities for Research in Astronomy
#                    (AURA)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials provided
#        with the distribution.
#
#     3. The name of AURA and its representatives may not be used to
#        endorse or promote products derived from this software without
#        specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS

from testtools import content

from pbr.tests import base


class TestCommands(base.BaseTestCase):
    def test_custom_build_py_command(self):
        """Test custom build_py command.

        Test that a custom subclass of the build_py command runs when listed in
        the commands [global] option, rather than the normal build command.
        """

        stdout, stderr, return_code = self.run_setup('build_py')
        self.addDetail('stdout', content.text_content(stdout))
        self.addDetail('stderr', content.text_content(stderr))
        self.assertIn('Running custom build_py command.', stdout)
        self.assertEqual(return_code, 0)

    def test_custom_rpm_version_py_command(self):
        """Test custom rpm_version command."""
        stdout, stderr, return_code = self.run_setup('rpm_version')
        self.addDetail('stdout', content.text_content(stdout))
        self.addDetail('stderr', content.text_content(stderr))
        self.assertIn('Extracting rpm version', stdout)
        self.assertEqual(return_code, 0)



# -*- coding: UTF-8 -*-
import unittest

from pandasticsearch.operators import *
from pandasticsearch.types import Row, Column


class TestSchema(unittest.TestCase):
    def test_row(self):
        row = Row(a=1, b='你好,世界')
        print(repr(row))

        self.assertEqual(row['a'], 1)
        self.assertEqual(row['b'], '你好,世界')
        self.assertEqual(row.as_dict(), {'a': 1, 'b': '你好,世界'})

    def test_column(self):
        col = Column('b')
        self._assert_equal_filter(col > 2, Greater('b', 2))
        self._assert_equal_filter(col >= 2, GreaterEqual('b', 2))
        self._assert_equal_filter(col < 2, Less('b', 2))
        self._assert_equal_filter(col <= 2, LessEqual('b', 2))
        self._assert_equal_filter(col == 2, Equal('b', 2))
        self._assert_equal_filter(col != 2, ~Equal('b', 2))
        self._assert_equal_filter(col.isin([1, 2, 3]), IsIn('b', [1, 2, 3]))
        self._assert_equal_filter(col.like('a*b'), Like('b', 'a*b'))
        self._assert_equal_filter(col.rlike('a*b'), Rlike('b', 'a*b'))
        self._assert_equal_filter(col.startswith('jj'), Startswith('b', 'jj'))
        self._assert_equal_filter(col.isnull, IsNull('b'))
        self._assert_equal_filter(col.notnull, NotNull('b'))

    def _assert_equal_filter(self, x, y):
        self.assertTrue(x, BooleanFilter)
        self.assertTrue(y, BooleanFilter)
        self.assertEqual(x.build(), y.build())


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2011 OpenStack Foundation
# Copyright (c) 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import print_function

import os
import sys
import tempfile
import testscenarios

try:
    import cStringIO as io
    BytesIO = io.StringIO
except ImportError:
    import io
    BytesIO = io.BytesIO

import fixtures

from pbr import git
from pbr import options
from pbr import packaging
from pbr.tests import base


class SkipFileWrites(base.BaseTestCase):

    scenarios = [
        ('changelog_option_true',
         dict(option_key='skip_changelog', option_value='True',
              env_key='SKIP_WRITE_GIT_CHANGELOG', env_value=None,
              pkg_func=git.write_git_changelog, filename='ChangeLog')),
        ('changelog_option_false',
         dict(option_key='skip_changelog', option_value='False',
              env_key='SKIP_WRITE_GIT_CHANGELOG', env_value=None,
              pkg_func=git.write_git_changelog, filename='ChangeLog')),
        ('changelog_env_true',
         dict(option_key='skip_changelog', option_value='False',
              env_key='SKIP_WRITE_GIT_CHANGELOG', env_value='True',
              pkg_func=git.write_git_changelog, filename='ChangeLog')),
        ('changelog_both_true',
         dict(option_key='skip_changelog', option_value='True',
              env_key='SKIP_WRITE_GIT_CHANGELOG', env_value='True',
              pkg_func=git.write_git_changelog, filename='ChangeLog')),
        ('authors_option_true',
         dict(option_key='skip_authors', option_value='True',
              env_key='SKIP_GENERATE_AUTHORS', env_value=None,
              pkg_func=git.generate_authors, filename='AUTHORS')),
        ('authors_option_false',
         dict(option_key='skip_authors', option_value='False',
              env_key='SKIP_GENERATE_AUTHORS', env_value=None,
              pkg_func=git.generate_authors, filename='AUTHORS')),
        ('authors_env_true',
         dict(option_key='skip_authors', option_value='False',
              env_key='SKIP_GENERATE_AUTHORS', env_value='True',
              pkg_func=git.generate_authors, filename='AUTHORS')),
        ('authors_both_true',
         dict(option_key='skip_authors', option_value='True',
              env_key='SKIP_GENERATE_AUTHORS', env_value='True',
              pkg_func=git.generate_authors, filename='AUTHORS')),
    ]

    def setUp(self):
        super(SkipFileWrites, self).setUp()
        self.temp_path = self.useFixture(fixtures.TempDir()).path
        self.root_dir = os.path.abspath(os.path.curdir)
        self.git_dir = os.path.join(self.root_dir, ".git")
        if not os.path.exists(self.git_dir):
            self.skipTest("%s is missing; skipping git-related checks"
                          % self.git_dir)
            return
        self.filename = os.path.join(self.temp_path, self.filename)
        self.option_dict = dict()
        if self.option_key is not None:
            self.option_dict[self.option_key] = ('setup.cfg',
                                                 self.option_value)
        self.useFixture(
            fixtures.EnvironmentVariable(self.env_key, self.env_value))

    def test_skip(self):
        self.pkg_func(git_dir=self.git_dir,
                      dest_dir=self.temp_path,
                      option_dict=self.option_dict)
        self.assertEqual(
            not os.path.exists(self.filename),
            (self.option_value.lower() in options.TRUE_VALUES
             or self.env_value is not None))

_changelog_content = """7780758\x00Break parser\x00 (tag: 1_foo.1)
04316fe\x00Make python\x00 (review/monty_taylor/27519)
378261a\x00Add an integration test script.\x00
3c373ac\x00Merge "Lib\x00 (HEAD, tag: 2013.2.rc2, tag: 2013.2, mile-proposed)
182feb3\x00Fix pip invocation for old versions of pip.\x00 (tag: 0.5.17)
fa4f46e\x00Remove explicit depend on distribute.\x00 (tag: 0.5.16)
d1c53dd\x00Use pip instead of easy_install for installation.\x00
a793ea1\x00Merge "Skip git-checkout related tests when .git is missing"\x00
6c27ce7\x00Skip git-checkout related tests when .git is missing\x00
451e513\x00Bug fix: create_stack() fails when waiting\x00
4c8cfe4\x00Improve test coverage: network delete API\x00 (tag: (evil))
d7e6167\x00Bug fix: Fix pass thru filtering in list_networks\x00 (tag: ev()il)
c47ec15\x00Consider 'in-use' a non-pending volume for caching\x00 (tag: ev)il)
8696fbd\x00Improve test coverage: private extension API\x00 (tag: ev(il)
f0440f8\x00Improve test coverage: hypervisor list\x00 (tag: e(vi)l)
04984a5\x00Refactor hooks file.\x00 (HEAD, tag: 0.6.7,b, tag: (12), master)
a65e8ee\x00Remove jinja pin.\x00 (tag: 0.5.14, tag: 0.5.13)
"""


class GitLogsTest(base.BaseTestCase):

    def setUp(self):
        super(GitLogsTest, self).setUp()
        self.temp_path = self.useFixture(fixtures.TempDir()).path
        self.root_dir = os.path.abspath(os.path.curdir)
        self.git_dir = os.path.join(self.root_dir, ".git")
        self.useFixture(
            fixtures.EnvironmentVariable('SKIP_GENERATE_AUTHORS'))
        self.useFixture(
            fixtures.EnvironmentVariable('SKIP_WRITE_GIT_CHANGELOG'))

    def test_write_git_changelog(self):
        self.useFixture(fixtures.FakePopen(lambda _: {
            "stdout": BytesIO(_changelog_content.encode('utf-8'))
        }))

        git.write_git_changelog(git_dir=self.git_dir,
                                dest_dir=self.temp_path)

        with open(os.path.join(self.temp_path, "ChangeLog"), "r") as ch_fh:
            changelog_contents = ch_fh.read()
            self.assertIn("2013.2", changelog_contents)
            self.assertIn("0.5.17", changelog_contents)
            self.assertIn("------", changelog_contents)
            self.assertIn("Refactor hooks file", changelog_contents)
            self.assertIn(
                "Bug fix: create_stack() fails when waiting",
                changelog_contents)
            self.assertNotIn("Refactor hooks file.", changelog_contents)
            self.assertNotIn("182feb3", changelog_contents)
            self.assertNotIn("review/monty_taylor/27519", changelog_contents)
            self.assertNotIn("0.5.13", changelog_contents)
            self.assertNotIn("0.6.7", changelog_contents)
            self.assertNotIn("12", changelog_contents)
            self.assertNotIn("(evil)", changelog_contents)
            self.assertNotIn("ev()il", changelog_contents)
            self.assertNotIn("ev(il", changelog_contents)
            self.assertNotIn("ev)il", changelog_contents)
            self.assertNotIn("e(vi)l", changelog_contents)
            self.assertNotIn('Merge "', changelog_contents)
            self.assertNotIn('1_foo.1', changelog_contents)

    def test_generate_authors(self):
        author_old = u"Foo Foo <email@foo.com>"
        author_new = u"Bar Bar <email@bar.com>"
        co_author = u"Foo Bar <foo@bar.com>"
        co_author_by = u"Co-authored-by: " + co_author

        git_log_cmd = (
            "git --git-dir=%s log --format=%%aN <%%aE>"
            % self.git_dir)
        git_co_log_cmd = ("git --git-dir=%s log" % self.git_dir)
        git_top_level = "git rev-parse --show-toplevel"
        cmd_map = {
            git_log_cmd: author_new,
            git_co_log_cmd: co_author_by,
            git_top_level: self.root_dir,
        }

        exist_files = [self.git_dir,
                       os.path.join(self.temp_path, "AUTHORS.in")]
        self.useFixture(fixtures.MonkeyPatch(
            "os.path.exists",
            lambda path: os.path.abspath(path) in exist_files))

        def _fake_run_shell_command(cmd, **kwargs):
            return cmd_map[" ".join(cmd)]

        self.useFixture(fixtures.MonkeyPatch(
            "pbr.git._run_shell_command",
            _fake_run_shell_command))

        with open(os.path.join(self.temp_path, "AUTHORS.in"), "w") as auth_fh:
            auth_fh.write("%s\n" % author_old)

        git.generate_authors(git_dir=self.git_dir,
                             dest_dir=self.temp_path)

        with open(os.path.join(self.temp_path, "AUTHORS"), "r") as auth_fh:
            authors = auth_fh.read()
            self.assertTrue(author_old in authors)
            self.assertTrue(author_new in authors)
            self.assertTrue(co_author in authors)


class BuildSphinxTest(base.BaseTestCase):

    scenarios = [
        ('true_autodoc_caps',
         dict(has_opt=True, autodoc='True', has_autodoc=True)),
        ('true_autodoc_caps_with_excludes',
         dict(has_opt=True, autodoc='True', has_autodoc=True,
              excludes="fake_package.fake_private_module\n"
              "fake_package.another_fake_*\n"
              "fake_package.unknown_module")),
        ('true_autodoc_lower',
         dict(has_opt=True, autodoc='true', has_autodoc=True)),
        ('false_autodoc',
         dict(has_opt=True, autodoc='False', has_autodoc=False)),
        ('no_autodoc',
         dict(has_opt=False, autodoc='False', has_autodoc=False)),
    ]

    def setUp(self):
        super(BuildSphinxTest, self).setUp()

        self.useFixture(fixtures.MonkeyPatch(
            "sphinx.setup_command.BuildDoc.run", lambda self: None))
        from distutils import dist
        self.distr = dist.Distribution()
        self.distr.packages = ("fake_package",)
        self.distr.command_options["build_sphinx"] = {
            "source_dir": ["a", "."]}
        pkg_fixture = fixtures.PythonPackage(
            "fake_package", [("fake_module.py", b""),
                             ("another_fake_module_for_testing.py", b""),
                             ("fake_private_module.py", b"")])
        self.useFixture(pkg_fixture)
        self.useFixture(base.DiveDir(pkg_fixture.base))
        self.distr.command_options["pbr"] = {}
        if hasattr(self, "excludes"):
            self.distr.command_options["pbr"]["autodoc_exclude_modules"] = (
                'setup.cfg',
                "fake_package.fake_private_module\n"
                "fake_package.another_fake_*\n"
                "fake_package.unknown_module")
        if self.has_opt:
            options = self.distr.command_options["pbr"]
            options["autodoc_index_modules"] = ('setup.cfg', self.autodoc)

    def test_build_doc(self):
        build_doc = packaging.LocalBuildDoc(self.distr)
        build_doc.run()

        self.assertTrue(
            os.path.exists("api/autoindex.rst") == self.has_autodoc)
        self.assertTrue(
            os.path.exists(
                "api/fake_package.fake_module.rst") == self.has_autodoc)
        if not self.has_autodoc or hasattr(self, "excludes"):
            assertion = self.assertFalse
        else:
            assertion = self.assertTrue
        assertion(
            os.path.exists(
                "api/fake_package.fake_private_module.rst"))
        assertion(
            os.path.exists(
                "api/fake_package.another_fake_module_for_testing.rst"))

    def test_builders_config(self):
        build_doc = packaging.LocalBuildDoc(self.distr)
        build_doc.finalize_options()

        self.assertEqual(2, len(build_doc.builders))
        self.assertIn('html', build_doc.builders)
        self.assertIn('man', build_doc.builders)

        build_doc = packaging.LocalBuildDoc(self.distr)
        build_doc.builders = ''
        build_doc.finalize_options()

        self.assertEqual('', build_doc.builders)

        build_doc = packaging.LocalBuildDoc(self.distr)
        build_doc.builders = 'man'
        build_doc.finalize_options()

        self.assertEqual(1, len(build_doc.builders))
        self.assertIn('man', build_doc.builders)

        build_doc = packaging.LocalBuildDoc(self.distr)
        build_doc.builders = 'html,man,doctest'
        build_doc.finalize_options()

        self.assertIn('html', build_doc.builders)
        self.assertIn('man', build_doc.builders)
        self.assertIn('doctest', build_doc.builders)

    def test_cmd_builder_override(self):

        if self.has_opt:
            self.distr.command_options["pbr"] = {
                "autodoc_index_modules": ('setup.cfg', self.autodoc)
            }

        self.distr.command_options["build_sphinx"]["builder"] = (
            "command line", "non-existing-builder")

        build_doc = packaging.LocalBuildDoc(self.distr)
        self.assertNotIn('non-existing-builder', build_doc.builders)
        self.assertIn('html', build_doc.builders)

        # process command line options which should override config
        build_doc.finalize_options()

        self.assertIn('non-existing-builder', build_doc.builders)
        self.assertNotIn('html', build_doc.builders)

    def test_cmd_builder_override_multiple_builders(self):

        if self.has_opt:
            self.distr.command_options["pbr"] = {
                "autodoc_index_modules": ('setup.cfg', self.autodoc)
            }

        self.distr.command_options["build_sphinx"]["builder"] = (
            "command line", "builder1,builder2")

        build_doc = packaging.LocalBuildDoc(self.distr)
        build_doc.finalize_options()

        self.assertEqual(["builder1", "builder2"], build_doc.builders)


class ParseRequirementsTestScenarios(base.BaseTestCase):

    versioned_scenarios = [
        ('non-versioned', {'versioned': False, 'expected': ['bar']}),
        ('versioned', {'versioned': True, 'expected': ['bar>=1.2.3']})
    ]

    scenarios = [
        ('normal', {'url': "foo\nbar", 'expected': ['foo', 'bar']}),
        ('normal_with_comments', {
            'url': "# this is a comment\nfoo\n# and another one\nbar",
            'expected': ['foo', 'bar']}),
        ('removes_index_lines', {'url': '-f foobar', 'expected': []}),
    ]

    scenarios = scenarios + testscenarios.multiply_scenarios([
        ('ssh_egg_url', {'url': 'git+ssh://foo.com/zipball#egg=bar'}),
        ('git_https_egg_url', {'url': 'git+https://foo.com/zipball#egg=bar'}),
        ('http_egg_url', {'url': 'https://foo.com/zipball#egg=bar'}),
    ], versioned_scenarios)

    scenarios = scenarios + testscenarios.multiply_scenarios(
        [
            ('git_egg_url',
                {'url': 'git://foo.com/zipball#egg=bar', 'name': 'bar'})
        ], [
            ('non-editable', {'editable': False}),
            ('editable', {'editable': True}),
        ],
        versioned_scenarios)

    def test_parse_requirements(self):
        tmp_file = tempfile.NamedTemporaryFile()
        req_string = self.url
        if hasattr(self, 'editable') and self.editable:
            req_string = ("-e %s" % req_string)
        if hasattr(self, 'versioned') and self.versioned:
            req_string = ("%s-1.2.3" % req_string)
        with open(tmp_file.name, 'w') as fh:
            fh.write(req_string)
        self.assertEqual(self.expected,
                         packaging.parse_requirements([tmp_file.name]))


class ParseRequirementsTest(base.BaseTestCase):

    def setUp(self):
        super(ParseRequirementsTest, self).setUp()
        (fd, self.tmp_file) = tempfile.mkstemp(prefix='openstack',
                                               suffix='.setup')

    def test_parse_requirements_override_with_env(self):
        with open(self.tmp_file, 'w') as fh:
            fh.write("foo\nbar")
        self.useFixture(
            fixtures.EnvironmentVariable('PBR_REQUIREMENTS_FILES',
                                         self.tmp_file))
        self.assertEqual(['foo', 'bar'],
                         packaging.parse_requirements())

    def test_parse_requirements_override_with_env_multiple_files(self):
        with open(self.tmp_file, 'w') as fh:
            fh.write("foo\nbar")
        self.useFixture(
            fixtures.EnvironmentVariable('PBR_REQUIREMENTS_FILES',
                                         "no-such-file," + self.tmp_file))
        self.assertEqual(['foo', 'bar'],
                         packaging.parse_requirements())

    def test_get_requirement_from_file_empty(self):
        actual = packaging.get_reqs_from_files([])
        self.assertEqual([], actual)

    def test_parse_requirements_python_version(self):
        with open("requirements-py%d.txt" % sys.version_info[0],
                  "w") as fh:
            fh.write("# this is a comment\nfoobar\n# and another one\nfoobaz")
        self.assertEqual(['foobar', 'foobaz'],
                         packaging.parse_requirements())

    def test_parse_requirements_right_python_version(self):
        with open("requirements-py1.txt", "w") as fh:
            fh.write("thisisatrap")
        with open("requirements-py%d.txt" % sys.version_info[0],
                  "w") as fh:
            fh.write("# this is a comment\nfoobar\n# and another one\nfoobaz")
        self.assertEqual(['foobar', 'foobaz'],
                         packaging.parse_requirements())


class ParseDependencyLinksTest(base.BaseTestCase):

    def setUp(self):
        super(ParseDependencyLinksTest, self).setUp()
        (fd, self.tmp_file) = tempfile.mkstemp(prefix="openstack",
                                               suffix=".setup")

    def test_parse_dependency_normal(self):
        with open(self.tmp_file, "w") as fh:
            fh.write("http://test.com\n")
        self.assertEqual(
            ["http://test.com"],
            packaging.parse_dependency_links([self.tmp_file]))

    def test_parse_dependency_with_git_egg_url(self):
        with open(self.tmp_file, "w") as fh:
            fh.write("-e git://foo.com/zipball#egg=bar")
        self.assertEqual(
            ["git://foo.com/zipball#egg=bar"],
            packaging.parse_dependency_links([self.tmp_file]))



# -*- coding: UTF-8 -*-

_metric_aggs = ('avg', 'min', 'max', 'cardinality', 'value_count', 'sum',
                'percentiles', 'percentile_ranks', 'stats', 'extended_stats')


class Aggregator(object):
    def __init__(self, field):
        self._field = field

    def build(self):
        pass


class MetricAggregator(Aggregator):
    def __init__(self, field, agg_type, alias=None, params=None):
        super(MetricAggregator, self).__init__(field)
        self._agg_type = agg_type
        self._alias = alias
        self._params = params

    def alias(self, alias):
        self._alias = alias
        return self

    def build(self):
        if self._agg_type not in _metric_aggs:
            raise Exception('Not support metric aggregator: {0}'.format(self._type))

        if self._alias is None:
            name = '{0}({1})'.format(self._agg_type, self._field)
        else:
            name = self._alias

        agg_field = dict()
        agg_field['field'] = self._field
        if self._params is not None:
            agg_field.update(self._params)
        return {name: {self._agg_type: agg_field}}




# -*- coding: UTF-8 -*-


# Es filter builder for BooleanCond
class BooleanFilter(object):
    def __init__(self, *args):
        self._filter = None

    def __and__(self, x):
        # Combine results
        if isinstance(self, AndFilter):
            self.subtree['must'].append(x.subtree)
            return self
        elif isinstance(x, AndFilter):
            x.subtree['must'].append(self.subtree)
            return x
        return AndFilter(self, x)

    def __or__(self, x):
        # Combine results
        if isinstance(self, OrFilter):
            self.subtree['should'].append(x.subtree)
            return self
        elif isinstance(x, OrFilter):
            x.subtree['should'].append(self.subtree)
            return x
        return OrFilter(self, x)

    def __invert__(self):
        return NotFilter(self)

    @property
    def subtree(self):
        if 'bool' in self._filter:
            return self._filter['bool']
        else:
            return self._filter

    def build(self):
        return self._filter


# Binary operator
class AndFilter(BooleanFilter):
    def __init__(self, *args):
        [isinstance(x, BooleanFilter) for x in args]
        super(AndFilter, self).__init__()
        self._filter = {'bool': {'must': [x.build() for x in args]}}


class OrFilter(BooleanFilter):
    def __init__(self, *args):
        [isinstance(x, BooleanFilter) for x in args]
        super(OrFilter, self).__init__()
        self._filter = {'bool': {'should': [x.build() for x in args]}}


class NotFilter(BooleanFilter):
    def __init__(self, x):
        assert isinstance(x, BooleanFilter)
        super(NotFilter, self).__init__()
        self._filter = {'bool': {'must_not': x.build()}}


# LeafBooleanFilter
class GreaterEqual(BooleanFilter):
    def __init__(self, field, value):
        super(GreaterEqual, self).__init__()
        self._filter = {'range': {field: {'gte': value}}}


class Greater(BooleanFilter):
    def __init__(self, field, value):
        super(Greater, self).__init__()
        self._filter = {'range': {field: {'gt': value}}}


class LessEqual(BooleanFilter):
    def __init__(self, field, value):
        super(LessEqual, self).__init__()
        self._filter = {'range': {field: {'lte': value}}}


class Less(BooleanFilter):
    def __init__(self, field, value):
        super(Less, self).__init__()
        self._filter = {'range': {field: {'lt': value}}}


class Equal(BooleanFilter):
    def __init__(self, field, value):
        super(Equal, self).__init__()
        self._filter = {'term': {field: value}}


class IsIn(BooleanFilter):
    def __init__(self, field, value):
        super(IsIn, self).__init__()
        assert isinstance(value, list)
        self._filter = {'terms': {field: value}}


class Like(BooleanFilter):
    def __init__(self, field, value):
        super(Like, self).__init__()
        self._filter = {'wildcard': {field: value}}


class Rlike(BooleanFilter):
    def __init__(self, field, value):
        super(Rlike, self).__init__()
        self._filter = {'regexp': {field: value}}


class Startswith(BooleanFilter):
    def __init__(self, field, value):
        super(Startswith, self).__init__()
        self._filter = {'prefix': {field: value}}


class IsNull(BooleanFilter):
    def __init__(self, field):
        super(IsNull, self).__init__()
        self._filter = {'missing': {'field': field}}


class NotNull(BooleanFilter):
    def __init__(self, field):
        super(NotNull, self).__init__()
        self._filter = {'exists': {'field': field}}


class ScriptFilter(BooleanFilter):
    def __init__(self, inline, lang=None, params=None):
        super(ScriptFilter, self).__init__()
        script = {'inline': inline}
        if lang is not None:
            script['lang'] = lang
        if params is not None:
            script['params'] = params
        self._filter = {'script': {'script': script}}


# -*- coding: UTF-8 -*-

import collections
import json
import six

from pandasticsearch.errors import NoSuchDependencyException


class Query(collections.MutableSequence):
    def __init__(self):
        super(Query, self).__init__()
        self._values = None
        self._result_dict = None
        self._took_millis = None

    def explain_result(self, result=None):
        if result is not None:
            assert isinstance(result, dict)
            self._result_dict = result
            self._took_millis = self._result_dict['took']

    def to_pandas(self):
        """
        Export the current query result to a Pandas DataFrame object.
        """
        raise NotImplementedError('implemented in subclass')

    def print_json(self):
        indented_json = json.dumps(self._result_dict, sort_keys=True, separators=(',', ': '), indent=4,
                                   ensure_ascii=False)
        print(indented_json)

    @property
    def result(self):
        return self._values

    @property
    def millis_taken(self):
        return self._took_millis

    @property
    def json(self):
        """
        Gets the original JSON representation returned by Elasticsearch REST API
        :return: The JSON string indicating the query result
        :rtype: string
        """
        return json.dumps(self._result_dict)

    def insert(self, index, value):
        self._values.insert(index, value)

    def append(self, value):
        self._values.append(value)

    def __str__(self):
        return str(self._values)

    def __len__(self):
        return len(self._values)

    def __delitem__(self, index):
        del self._values[index]

    def __setitem__(self, index, value):
        self._values[index] = value

    def __getitem__(self, index):
        return self._values[index]


class Select(Query):
    def __init__(self):
        super(Select, self).__init__()

    def resolve_fields(self, row):
        fields = {}
        for field in row:
            nested_fields = {}
            if isinstance(row[field], dict):
                nested_fields = self.resolve_fields(row[field])
                for n_field, val in nested_fields.items():
                    fields["{}.{}".format(field, n_field)] = val
            else:
                fields[field] = row[field]
        return fields

    def explain_result(self, result=None):
        super(Select, self).explain_result(result)
        rows = []
        for hit in self._result_dict['hits']['hits']:
            row = {}
            for k in hit.keys():
                if k == '_source':
                    solved_fields = self.resolve_fields(hit['_source'])
                    row.update(solved_fields)
                elif k.startswith('_'):
                    row[k] = hit[k]
            rows.append(row)
        self._values = rows

    def to_pandas(self):
        try:
            import pandas
        except ImportError:
            raise NoSuchDependencyException('this method requires pandas library')
        if self._values:
            df = pandas.DataFrame(data=self._values)
            return df

    @staticmethod
    def from_dict(d):
        query = Select()
        query.explain_result(d)
        return query

    @classmethod
    def _stringfy_value(cls, value):
        b = six.StringIO()
        if value:
            b.write(repr(value))
        else:
            b.write('(NULL)')
        return b.getvalue()

    def result_as_tabular(self, cols, n, truncate=20):
        b = six.StringIO()
        widths = []
        tavnit = '|'
        separator = '+'

        for col in cols:
            maxlen = len(col)
            for kv in self.result[:n]:
                if col in kv:
                    s = Select._stringfy_value(kv[col])
                else:
                    s = '(NULL)'
                if len(s) > maxlen:
                    maxlen = len(s)
            widths.append(min(maxlen, truncate))

        for w in widths:
            tavnit += ' %-' + '%ss |' % (w,)
            separator += '-' * w + '--+'

        b.write(separator + '\n')
        b.write(tavnit % tuple(cols) + '\n')
        b.write(separator + '\n')
        for kv in self.result[:n]:
            row = []
            for col in cols:
                if col in kv:
                    s = Select._stringfy_value(kv[col])
                    if len(s) > truncate:
                        s = s[:truncate - 3] + '...'
                else:
                    s = '(NULL)'
                row.append(s)
            b.write(tavnit % tuple(row) + '\n')
        b.write(separator + '\n')
        return b.getvalue()


class Agg(Query):
    def __init__(self):
        super(Agg, self).__init__()
        self._index_names = None
        self._indexes = None

    def explain_result(self, result=None):
        super(Agg, self).explain_result(result)
        tuples = list(Agg._process_agg(self._result_dict['aggregations']))
        assert len(tuples) > 0
        self._index_names = list(tuples[0][0])
        self._values = []
        self._indexes = []
        for t in tuples:
            _, index, row = t
            self._values.append(row)
            if len(index) > 0:
                self._indexes.append(index)

    @property
    def index(self):
        return self._indexes

    def to_pandas(self):
        try:
            import pandas
        except ImportError:
            raise NoSuchDependencyException('this method requires pandas library')
        if self._values is not None:
            if len(self._indexes) > 0:
                index = pandas.MultiIndex.from_tuples(self._indexes, names=self._index_names)
                df = pandas.DataFrame(data=self._values, index=index)
            else:
                df = pandas.DataFrame(data=self._values)
            return df

    @classmethod
    def _process_agg(cls, bucket, indexes=(), names=()):
        """
        Recursively extract agg values
        :param bucket: a bucket contains either sub-buckets or a bunch of aggregated values
        :return: a list of tuples: (index_name, index_tuple, row)
        """
        # for each agg, yield a row
        row = {}
        for k, v in bucket.items():
            if isinstance(v, dict):
                if 'buckets' in v:
                    for sub_bucket in v['buckets']:

                        if 'key_as_string' in sub_bucket:
                            key = sub_bucket['key_as_string']
                        else:
                            key = sub_bucket['key']
                        for x in Agg._process_agg(sub_bucket,
                                                  indexes + (key,),
                                                  names + (k,)):
                            yield x
                elif 'value' in v:
                    row[k] = v['value']
                elif 'values' in v:  # percentiles
                    row = v['values']
                else:
                    row.update(v)  # stats
            else:
                if k == 'doc_count':  # count docs
                    row['doc_count'] = v

        if len(row) > 0:
            yield (names, indexes, row)

    @staticmethod
    def from_dict(d):
        agg = Agg()
        agg.explain_result(d)
        return agg



# -*- coding: UTF-8 -*-

from pandasticsearch.operators.aggregator import *
from pandasticsearch.operators.grouper import *
from pandasticsearch.operators.filter import *
from pandasticsearch.operators.sorter import *



# -*- coding: UTF-8 -*-

from pandasticsearch.operators import *
import six


class Column(object):
    def __init__(self, field):
        self._field = field

    def field_name(self):
        return self._field

    def __eq__(self, other):
        return Equal(field=self._field, value=other)

    def __ne__(self, other):
        return ~Equal(field=self._field, value=other)

    def __gt__(self, other):
        return Greater(field=self._field, value=other)

    def __lt__(self, other):
        return Less(field=self._field, value=other)

    def __ge__(self, other):
        return GreaterEqual(field=self._field, value=other)

    def __le__(self, other):
        return LessEqual(field=self._field, value=other)

    def isin(self, values):
        """
        Returns a :class:`BooleanFilter <pandasticsearch.operators.BooleanFilter>`

        :param values:  A list of values to filter terms
        :return: :class:`BooleanFilter <pandasticsearch.operators.BooleanFilter>`

        df.filter(df.gender.isin(['male', 'female'])
        """
        return IsIn(field=self._field, value=values)

    def like(self, wildcard):
        """
        Returns a :class:`BooleanFilter <pandasticsearch.operators.BooleanFilter>`

        :param str wildcard: The wildcard to filter the column with.
        :return: :class:`BooleanFilter <pandasticsearch.operators.BooleanFilter>`

        >>> df.filter(df.name.like('A*'))
        """
        return Like(field=self._field, value=wildcard)

    def rlike(self, regexp):
        """
        Returns a :class:`BooleanFilter <pandasticsearch.operators.BooleanFilter>`

        :param str regexp: The regular expression to filter the column with.
        :return: :class:`BooleanFilter <pandasticsearch.operators.BooleanFilter>`

        >>> df.filter(df.name.rlike('A.l.e'))
        """
        return Rlike(field=self._field, value=regexp)

    def startswith(self, substr):
        """
        Returns a :class:`BooleanFilter <pandasticsearch.operators.BooleanFilter>`

        :param str substr: The sub string to filter the column with.
        :return: :class:`BooleanFilter <pandasticsearch.operators.BooleanFilter>`

        >>> df.filter(df.name.startswith('Al')
        """
        return Startswith(field=self._field, value=substr)

    def ranges(self, values):
        """
        Returns a :class:`Grouper <pandasticsearch.operators.Grouper>`

        :param values: A list of numeric values
        :return: :class:`Grouper <pandasticsearch.operators.Grouper>`

        >>> df.groupby(df.age.ranges([10,12,14]))
        """
        return RangeGrouper(field=self._field, range_list=values)

    def date_interval(self, interval, format='yyyy/MM/dd HH:mm:ss'):
        """
        Returns a :class:`Grouper <pandasticsearch.operators.Grouper>`

        :param interval: A string indicating date interval
        :param format: Date format string
        :return: :class:`Grouper <pandasticsearch.operators.Grouper>`

        >>> df.groupby(df.date_interval('1d'))
        """
        return DateGrouper(field=self._field, interval=interval, format=format)

    def terms(self, limit=20, include=None, exclude=None):
        """
        Returns a :class:`Grouper <pandasticsearch.operators.Grouper>`

        :param limit: limit the number of terms to be aggregated (default 20)
        :param include: the exact term to be included
        :param exclude: the exact term to be excluded

        :return: :class:`Grouper <pandasticsearch.operators.Grouper>`

        >>> df.groupby(df.age.terms(limit=10, include=[1, 2, 3]))
        """
        return Grouper(field=self._field, size=limit, include=include, exclude=exclude)

    @property
    def isnull(self):
        """
        :class:`BooleanFilter <pandasticsearch.operators.BooleanFilter>` to indicate the null column value

        :return: :class:`BooleanFilter <pandasticsearch.operators.BooleanFilter>`
        """
        return IsNull(field=self._field)

    @property
    def notnull(self):
        """
        :class:`BooleanFilter <pandasticsearch.operators.BooleanFilter>` to indicate the non-null column value

        :return: :class:`BooleanFilter <pandasticsearch.operators.BooleanFilter>`
        """
        return NotNull(field=self._field)

    @property
    def desc(self, mode=None):
        """
        Descending :class:`Sorter <pandasticsearch.operators.Sorter>`

        :return: :class:`Sorter <pandasticsearch.operators.Sorter>`

        >>> df.orderyby(df.age.desc)
        """
        return Sorter(self._field, mode=mode)

    @property
    def asc(self, mode=None):
        """
        Ascending :class:`Sorter <pandasticsearch.operators.Sorter>`

        :return: :class:`Sorter <pandasticsearch.operators.Sorter>`

        >>> df.orderyby(df.age.asc)
        """
        return Sorter(self._field, order='asc', mode=mode)

    @property
    def max(self):
        """
        Max aggregator

        :return: :class:`Aggregator <pandasticsearch.operators.Aggregator>`

        >>> df.groupby(df.gender).agg(df.age.max)
        """
        return MetricAggregator(self._field, 'max')

    @property
    def min(self):
        """
        Min aggregator

        :return: :class:`Aggregator <pandasticsearch.operators.Aggregator>`

        >>> df.groupby(df.gender).agg(df.age.min)
        """
        return MetricAggregator(self._field, 'min')

    @property
    def avg(self):
        """
        Avg aggregator

        :return: :class:`Aggregator <pandasticsearch.operators.Aggregator>`

        >>> df.groupby(df.gender).agg(df.age.avg)
        """
        return MetricAggregator(self._field, 'avg')

    @property
    def sum(self):
        """
        Sum aggregator

        :return: :class:`Aggregator <pandasticsearch.operators.Aggregator>`

        >>> df.groupby(df.gender).agg(df.age.sum)
        """
        return MetricAggregator(self._field, 'sum')

    @property
    def value_count(self):
        """
        Value count aggregator

        :return: :class:`Aggregator <pandasticsearch.operators.Aggregator>`

        >>> df.groupby(df.gender).agg(df.age.value_count)
        """
        return MetricAggregator(self._field, 'value_count')

    count = value_count

    @property
    def cardinality(self):
        """
        Distince aggregator

        :return: :class:`Aggregator <pandasticsearch.operators.Aggregator>`

        >>> df.groupby(df.gender).agg(df.age.cardinality)
        >>> df.groupby(df.gender).agg(df.age.distinct_count)
        """
        return MetricAggregator(self._field, 'cardinality')

    distinct_count = cardinality

    @property
    def percentiles(self):
        """
        Percentile aggregator

        :return: :class:`Aggregator <pandasticsearch.operators.Aggregator>`

        >>> df.groupby(df.gender).agg(df.age.percentiles)
        """
        return MetricAggregator(self._field, 'percentiles')

    @property
    def percentile_ranks(self):
        """
        Percentile ranks aggregator

        :return: :class:`Aggregator <pandasticsearch.operators.Aggregator>`

        >>> df.groupby(df.gender).agg(df.age.percentile_ranks)
        """
        return MetricAggregator(self._field, 'percentile_ranks')

    @property
    def stats(self):
        """
        Stats aggregator

        :return: :class:`Aggregator <pandasticsearch.operators.Aggregator>`

        >>> df.groupby(df.gender).agg(df.age.stats)
        """
        return MetricAggregator(self._field, 'stats')

    @property
    def extended_stats(self):
        """
        Extended stats aggregator

        :return: :class:`Aggregator <pandasticsearch.operators.Aggregator>`

        >>> df.groupby(df.gender).agg(df.age.extended_stats)
        """
        return MetricAggregator(self._field, 'extended_stats')


class Row(tuple):
    """
    The builtin :class:`DataFrame <pandasticsearch.dataframe.DataFrame>` row type for accessing before converted into Pandas DataFrame.
    The fields will be sorted by names.

    >>> row = Row(name="Alice", age=12)
    >>> row
    Row(age=12, name='Alice')
    >>> row['name'], row['age']
    ('Alice', 12)
    >>> row.name, row.age
    ('Alice', 12)
    >>> 'name' in row
    True
    >>> 'wrong_key' in row
    """

    def __new__(cls, **kwargs):
        names = sorted(kwargs.keys())
        row = tuple.__new__(cls, [kwargs[n] for n in names])
        row._fields = names
        return row

    def __getitem__(self, name):
        try:
            idx = self._fields.index(name)
            return super(Row, self).__getitem__(idx)
        except IndexError:
            raise KeyError(name)
        except ValueError:
            raise ValueError(name)

    def __contains__(self, name):
        return name in self._fields

    def __repr__(self):
        return 'Row(' + ','.join(
            ['{0}={1}'.format(k, Row._stringfy(v)) for k, v in zip(self._fields, tuple(self))]) + ')'

    @classmethod
    def _stringfy(cls, v):
        b = six.StringIO()
        b.write(repr(v))
        return b.getvalue()

    def as_dict(self):
        return dict((x, y) for x, y in zip(self._fields, self))



# -*- coding: UTF-8 -*-


_sort_mode = ('min', 'max', 'sum', 'avg', 'median')


class Sorter(object):
    def __init__(self, field, order='desc', mode=None):
        self._field = field
        self._order = order
        self._mode = mode

    def build(self):
        sort = {}
        if self._mode is not None:
            if self._mode not in _sort_mode:
                raise Exception('Not support sort mode: {0}'.format(self._mode))
            sort['mode'] = self._mode
        sort['order'] = self._order
        return {self._field: sort}


class ScriptSorter(object):
    def __init__(self, script, order='desc', type='number', params=None):
        self._order = order
        self._script = script
        self._params = params
        self._type = type

    def build(self):
        script = {'script': self._script, 'type': self._type, 'order': self._order}
        if self._params:
            script['params'] = self._params
        return {'_script': script}





from __future__ import absolute_import

from pandasticsearch.dataframe import DataFrame, Column
from pandasticsearch.client import RestClient
from pandasticsearch.queries import Select, Agg
from pandasticsearch.types import Row

col = Column



# -*- coding: UTF-8 -*-
import unittest

from pandasticsearch.operators import *


class TestOperators(unittest.TestCase):
    def test_metric_agg(self):
        self.assertEqual(MetricAggregator('x', 'avg').build(), {'avg(x)': {'avg': {'field': 'x'}}})
        self.assertEqual(MetricAggregator('x', 'max').build(), {'max(x)': {'max': {'field': 'x'}}})
        self.assertEqual(MetricAggregator('x', 'max').alias('max_x').build(), {'max_x': {'max': {'field': 'x'}}})

    def test_grouper(self):
        exp = Grouper('a', size=100, include=['x', 'y'])
        self.assertEqual(exp.build(),
                         {'a': {'terms': {'field': 'a', 'size': 100, 'include': ['x', 'y']}}})

    def test_nested_grouper(self):
        exp = Grouper('a', inner=Grouper('b', inner=Grouper('c')))
        self.assertEqual(exp.build(),
                         {
                             'a': {
                                 'terms': {'field': 'a', 'size': 20},
                                 'aggregations': {
                                     'b': {
                                         'terms': {'field': 'b', 'size': 20},
                                         'aggregations': {
                                             'c': {'terms': {'field': 'c', 'size': 20}}}}}}})

    def test_range_grouper(self):
        exp = RangeGrouper('a', [1, 3, 6])
        self.assertEqual(exp.build(), {
            'range(1,3,6)': {'range': {'ranges': [{'to': 3, 'from': 1}, {'to': 6, 'from': 3}], 'field': 'a'}}})

    def test_date_grouper(self):
        exp = DateGrouper('a', '1d', 'm')
        self.assertEqual(exp.build(), {
            'date(a,1d)': {'date_histogram': {'interval': '1d', 'field': 'a', 'format': 'm'}}})

    def test_sorter(self):
        self.assertEqual(Sorter('x').build(), {'x': {'order': 'desc'}})
        self.assertEqual(Sorter('x', mode='avg').build(), {'x': {'order': 'desc', 'mode': 'avg'}})

        self.assertEqual(ScriptSorter('doc["field_name"].value * factor', params={'factor': 1.1}).build(),
                         {'_script': {'type': 'number',
                                      'order': 'desc',
                                      'script': 'doc["field_name"].value * factor',
                                      "params": {
                                          "factor": 1.1
                                      }}})

    def test_leaf_boolean_filter(self):
        self.assertEqual(GreaterEqual('a', 2).build(), {"range": {"a": {"gte": 2}}})
        self.assertEqual(LessEqual('a', 2).build(), {"range": {"a": {"lte": 2}}})
        self.assertEqual(Less('a', 2).build(), {"range": {"a": {"lt": 2}}})
        self.assertEqual(Equal('a', 2).build(), {"term": {"a": 2}})
        exp = Equal('a', 2)
        self.assertEqual((~exp).build()['bool'], {"must_not": {"term": {"a": 2}}})
        self.assertEqual(Greater('a', 2).build(), {"range": {"a": {"gt": 2}}})
        self.assertEqual(IsIn('a', [1, 2, 3]).build(), {'terms': {'a': [1, 2, 3]}})
        self.assertEqual(Like('a', 'a*b').build(), {'wildcard': {'a': 'a*b'}})
        self.assertEqual(Rlike('a', 'a*b').build(), {'regexp': {'a': 'a*b'}})
        self.assertEqual(Startswith('a', 'jj').build(), {'prefix': {'a': 'jj'}})
        self.assertEqual(IsNull('a').build(), {'missing': {'field': 'a'}})
        self.assertEqual(NotNull('a').build(), {'exists': {'field': 'a'}})
        self.assertEqual(ScriptFilter('doc["num1"].value > params.param1', params={'param1': 5}).build(),
                         {'script': {
                             'script': {
                                 'inline': 'doc["num1"].value > params.param1',
                                 'params': {'param1': 5}}}})

    def test_and_filter1(self):
        exp = GreaterEqual('a', 2) & Less('b', 3)
        self.assertEqual(
            exp.build(),
            {
                'bool': {
                    'must': [
                        {'range': {'a': {'gte': 2}}},
                        {'range': {'b': {'lt': 3}}}
                    ]
                }
            })

    def test_and_filter2(self):
        exp = GreaterEqual('a', 2) & Less('b', 3) & Equal('c', 4)
        self.assertEqual(
            exp.build(),
            {
                'bool': {
                    'must': [
                        {'range': {'a': {'gte': 2}}},
                        {'range': {'b': {'lt': 3}}},
                        {'term': {'c': 4}}
                    ]
                }
            })

    def test_and_filter3(self):
        exp = GreaterEqual('a', 2) & (Less('b', 3) & Equal('c', 4))
        self.assertEqual(
            exp.build(),
            {
                'bool': {
                    'must': [
                        {'range': {'b': {'lt': 3}}},
                        {'term': {'c': 4}},
                        {'range': {'a': {'gte': 2}}}
                    ]
                }
            })

    def test_or_filter1(self):
        exp = GreaterEqual('a', 2) | Less('b', 3)
        self.assertEqual(
            exp.build(),
            {
                'bool': {
                    'should': [
                        {'range': {'a': {'gte': 2}}},
                        {'range': {'b': {'lt': 3}}}
                    ]
                }
            })

    def test_or_filter2(self):
        exp = GreaterEqual('a', 2) | Less('b', 3) | Equal('c', 4)
        self.assertEqual(
            exp.build(),
            {
                'bool': {
                    'should': [
                        {'range': {'a': {'gte': 2}}},
                        {'range': {'b': {'lt': 3}}},
                        {'term': {'c': 4}}
                    ]
                }
            })

    def test_or_filter3(self):
        exp = GreaterEqual('a', 2) | (Less('b', 3) | Equal('c', 4))
        self.assertEqual(
            exp.build(),
            {
                'bool': {
                    'should': [
                        {'range': {'b': {'lt': 3}}},
                        {'term': {'c': 4}},
                        {'range': {'a': {'gte': 2}}}
                    ]
                }
            })

    def test_not_filter(self):
        exp = ~GreaterEqual('a', 2)
        self.assertEqual(
            exp.build(),
            {
                'bool': {
                    'must_not': {'range': {'a': {'gte': 2}}}
                }
            })

    def test_not_not_filter(self):
        exp = ~~GreaterEqual('a', 2)

        self.assertEqual(
            exp.build(),
            {
                'bool': {
                    'must_not': {
                        'bool': {
                            'must_not': {'range': {'a': {'gte': 2}}}
                        }
                    }
                }
            })

    def test_not_and_filter(self):
        exp = ~(GreaterEqual('a', 2) & Less('b', 3))
        self.assertEqual(
            exp.build(),
            {
                'bool': {
                    'must_not': {
                        'bool': {
                            'must': [
                                {'range': {'a': {'gte': 2}}},
                                {'range': {'b': {'lt': 3}}}
                            ]
                        }
                    }
                }
            })

    def test_and_or_filter(self):
        exp = GreaterEqual('a', 2) & (Less('b', 3) | Equal('c', 4))
        self.assertEqual(
            exp.build(),
            {
                'bool': {
                    'must': [
                        {'range': {'a': {'gte': 2}}},
                        {
                            'bool': {
                                'should': [
                                    {'range': {'b': {'lt': 3}}},
                                    {'term': {'c': 4}}
                                ]
                            }
                        }
                    ]
                }
            })

    def test_and_not_or_filter(self):
        exp = GreaterEqual('a', 2) & ~(Less('b', 3) | Equal('c', 4))
        self.assertEqual(
            exp.build(),
            {
                'bool': {
                    'must': [
                        {'range': {'a': {'gte': 2}}},
                        {
                            'bool': {
                                'must_not': {
                                    'bool': {
                                        'should': [
                                            {'range': {'b': {'lt': 3}}},
                                            {'term': {'c': 4}}
                                        ]
                                    }

                                }
                            }
                        }
                    ]
                }
            })


if __name__ == '__main__':
    unittest.main()



from setuptools import setup

with open('requirements.txt') as f:
    required = f.read().splitlines()

extras_require = {
    "pandas": ["pandas"],
}

setup(
    name='pandasticsearch',
    version='0.5.3',
    author='onesuper',
    author_email='onesuperclark@gmail.com',
    packages=['pandasticsearch', 'pandasticsearch.operators'],
    url='http://pypi.python.org/pypi/pandasticsearch/',
    license='MIT',
    description='A Pandastic Elasticsearch client for data analyzing.',
    install_requires=required,
    extras_require=extras_require,
    test_suite='nose.collector',
    tests_require=['nose', 'mock'],
)



# -*- coding: UTF-8 -*-

from pandasticsearch.client import RestClient
from pandasticsearch.queries import Agg, Select
from pandasticsearch.operators import *
from pandasticsearch.types import Column, Row
from pandasticsearch.errors import DataFrameException

import json
import six
import sys
import copy

_unbound_index_err = DataFrameException('DataFrame is not bound to ES index')

_count_aggregator = MetricAggregator('_index', 'value_count', alias='count').build()


class DataFrame(object):
    """
    A :class:`DataFrame` treats index and documents in Elasticsearch as named columns and rows.

    >>> from pandasticsearch import DataFrame
    >>> df = DataFrame.from_es('http://localhost:9200', index='people')

    Customizing the endpoint of the ElasticSearch:

    >>> from pandasticsearch import DataFrame
    >>> from pandasticsearch.client import RestClient
    >>> df = DataFrame(client=RestClient('http://host:port/v2/_search',), index='people')

    It can be converted to Pandas object for subsequent analysis:

    >>> df.to_pandas()
    """

    def __init__(self, **kwargs):
        self._client = kwargs.get('client', None)
        self._mapping = kwargs.get('mapping', None)
        self._doc_type = kwargs.get('doc_type', None)
        self._index = kwargs.get('index', None)
        self._compat = kwargs.get('compat', 2)
        self._filter = kwargs.get('filter', None)
        self._groupby = kwargs.get('groupby', None)
        self._aggregation = kwargs.get('aggregation', None)
        self._sort = kwargs.get('sort', None)
        self._projection = kwargs.get('projection', None)
        self._limit = kwargs.get('limit', None)
        self._last_query = None

    @property
    def index(self):
        """
        Returns the index name.

        :return: string as the name

        >>> df.index
        people/children
        """
        if self._index is None:
            return None
        return self._index + '/' + self._doc_type if self._doc_type else self._index

    @property
    def columns(self):
        """
        Returns all column names as a list.

        :return: column names as a list

        >>> df.columns
        ['age', 'name']
        """
        return sorted(self._get_cols(self._mapping)) if self._mapping else None

    @property
    def schema(self):
        """
        Returns the schema(mapping) of the index/type as a dictionary.
        """
        return self._mapping

    @staticmethod
    def from_es(**kwargs):
        """
        Creates an :class:`DataFrame <DataFrame>` object by providing the URL of ElasticSearch node and the name of the index.

        :param str url: URL of the node connected to (default: 'http://localhost:9200')
        :param str index: The name of the index
        :param str doc_type: The type of the document
        :param str compat: The compatible ES version (an integer number)
        :return: DataFrame object for accessing
        :rtype: DataFrame

        >>> from pandasticsearch import DataFrame
        >>> df = DataFrame.from_es('http://localhost:9200', index='people')
        """

        doc_type = kwargs.get('doc_type', None)
        index = kwargs.get('index', None)
        url = kwargs.get('url', 'http://localhost:9200')
        compat = kwargs.get('compat', 2)
        username = kwargs.get('username', None)
        password = kwargs.get('password', None)
        verify_ssl = kwargs.get('verify_ssl', True)

        if index is None:
            raise ValueError('Index name must be specified')

        mapping = RestClient(url, index, username, password, verify_ssl).get()

        if doc_type is None:
            endpoint = index + '/_search'
        else:
            endpoint = index + '/' + doc_type + '/_search'
        return DataFrame(client=RestClient(url, endpoint, username, password, verify_ssl),
                         mapping=mapping, index=index, doc_type=doc_type, compat=compat)

    def __getattr__(self, name):
        """
        Returns a :class:`types.Column <pandasticsearch.types.Column>` object denoted by ``name``.
        """
        if name not in self.columns:
            raise AttributeError(
                "'%s' object has no attribute '%s'" % (self.__class__.__name__, name))
        return Column(name)

    def __getitem__(self, item):
        if isinstance(item, six.string_types):
            if item not in self.columns:
                raise TypeError('Column does not exist: [{0}]'.format(item))
            return Column(item)
        elif isinstance(item, BooleanFilter):
            self._filter = item
            return self
        else:
            raise TypeError('Unsupported expr: [{0}]'.format(item))

    def filter(self, condition):
        """
        Filters rows using a given condition.

        where() is an alias for filter().

        :param condition: :class:`BooleanFilter <pandasticsearch.operators.BooleanFilter>` object or a string

        >>> df.filter(df['age'] < 13).collect()
        [Row(age=12,gender='female',name='Alice'), Row(age=11,gender='male',name='Bob')]
        """

        if isinstance(condition, six.string_types):
            _filter = ScriptFilter(condition)
        elif isinstance(condition, BooleanFilter):
            _filter = condition
        else:
            raise TypeError('{0} is supposed to be str or BooleanFilter'.format(condition))

        # chaining filter treated as AND
        if self._filter is not None:
            _filter = (self._filter & _filter)

        return DataFrame(client=self._client,
                         index=self._index,
                         doc_type=self._doc_type,
                         mapping=self._mapping,
                         filter=_filter,
                         groupby=self._groupby,
                         aggregation=self._aggregation,
                         projection=self._projection,
                         sort=self._sort,
                         limit=self._limit,
                         compat=self._compat)

    where = filter

    def select(self, *cols):
        """
        Projects a set of columns and returns a new :class:`DataFrame <DataFrame>`

        :param cols: list of column names or :class:`Column <pandasticsearch.types.Column>`.

        >>> df.filter(df['age'] < 25).select('name', 'age').collect()
        [Row(age=12,name='Alice'), Row(age=11,name='Bob'), Row(age=13,name='Leo')]
        """
        projection = []
        for col in cols:
            if isinstance(col, six.string_types):
                projection.append(getattr(self, col))
            elif isinstance(col, Column):
                projection.append(col)
            else:
                raise TypeError('{0} is supposed to be str or Column'.format(col))
        return DataFrame(client=self._client,
                         index=self._index,
                         doc_type=self._doc_type,
                         mapping=self._mapping,
                         filter=self._filter,
                         groupby=self._groupby,
                         aggregation=self._aggregation,
                         projection=projection,
                         sort=self._sort,
                         limit=self._limit,
                         compat=self._compat)

    def limit(self, num):
        """
        Limits the result count to the number specified.
        """
        assert isinstance(num, int)
        assert num >= 1
        return DataFrame(client=self._client,
                         index=self._index,
                         doc_type=self._doc_type,
                         mapping=self._mapping,
                         filter=self._filter,
                         groupby=self._groupby,
                         aggregation=self._aggregation,
                         projection=self._projection,
                         sort=self._sort,
                         limit=num,
                         compat=self._compat)

    def groupby(self, *cols):
        """
        Returns a new :class:`DataFrame <DataFrame>` object grouped by the specified column(s).

        :param cols: A list of column names, :class:`Column <pandasticsearch.types.Column>` or :class:`Grouper <pandasticsearch.operators.Grouper>` objects
        """
        columns = []
        if len(cols) == 1 and isinstance(cols[0], Grouper):
            groupby = cols[0].build()
        else:
            for col in cols:
                if isinstance(col, six.string_types):
                    columns.append(getattr(self, col))
                elif isinstance(col, Column):
                    columns.append(col)
                else:
                    raise TypeError('{0} is supposed to be str or Column'.format(col))
            names = [col.field_name() for col in columns]
            groupby = Grouper.from_list(names).build()

        return DataFrame(client=self._client,
                         index=self._index,
                         doc_type=self._doc_type,
                         mapping=self._mapping,
                         filter=self._filter,
                         groupby=groupby,
                         aggregation=self._aggregation,
                         projection=self._projection,
                         sort=self._sort,
                         limit=self.limit,
                         compat=self._compat)

    def agg(self, *aggs):
        """
        Aggregate on the entire DataFrame without groups.

        :param aggs: a list of :class:`Aggregator <pandasticsearch.operators.Aggregator>` objects

        >>> df[df['gender'] == 'male'].agg(df['age'].avg).collect()
        [Row(avg(age)=12)]
        """
        aggregation = {}
        for agg in aggs:
            assert isinstance(agg, Aggregator)
            aggregation.update(agg.build())

        return DataFrame(client=self._client,
                         index=self._index,
                         doc_type=self._doc_type,
                         mapping=self._mapping,
                         filter=self._filter,
                         groupby=self._groupby,
                         aggregation=aggregation,
                         projection=self._projection,
                         sort=self._sort,
                         limit=self._limit,
                         compat=self._compat)

    def sort(self, *cols):
        """
        Returns a new :class:`DataFrame <DataFrame>` object sorted by the specified column(s).

        :param cols: A list of column names, :class:`Column <pandasticsearch.types.Column>` or :class:`Sorter <pandasticsearch.operators.Sorter>`.

        orderby() is an alias for sort().

        >>> df.sort(df['age'].asc).collect()
        [Row(age=11,name='Bob'), Row(age=12,name='Alice'), Row(age=13,name='Leo')]
        """
        sorts = []
        for col in cols:
            if isinstance(col, six.string_types):
                sorts.append(ScriptSorter(col).build())
            elif isinstance(col, Sorter):
                sorts.append(col.build())
            else:
                raise TypeError('{0} is supposed to be str or Sorter'.format(col))

        return DataFrame(client=self._client,
                         index=self._index,
                         doc_type=self._doc_type,
                         mapping=self._mapping,
                         filter=self._filter,
                         groupby=self._groupby,
                         aggregation=self._aggregation,
                         projection=self._projection,
                         sort=sorts,
                         limit=self._limit,
                         compat=self._compat)

    orderby = sort

    def _execute(self):
        if self._client is None:
            raise _unbound_index_err

        res_dict = self._client.post(data=self._build_query())
        if self._aggregation is None and self._groupby is None:
            query = Select()
            query.explain_result(res_dict)
        else:
            query = Agg.from_dict(res_dict)
        return query

    def collect(self):
        """
        Returns all the records as a list of Row.

        :return: list of :class:`Row <pandasticsearch.types.Row>`

        >>> df.collect()
        [Row(age=2, name='Alice'), Row(age=5, name='Bob')]
        """
        query = self._execute()
        return [Row(**v) for v in query.result]

    def to_pandas(self):
        """
        Export to a Pandas DataFrame object.

        :return: The DataFrame representing the query result

        >>> df[df['gender'] == 'male'].agg(Avg('age')).to_pandas()
            avg(age)
        0        12
        """
        query = self._execute()
        return query.to_pandas()

    def count(self):
        """
        Returns a list of numbers indicating the count for each group

        >>> df.groupby(df.gender).count()
        [2, 1]
        """
        df = DataFrame(client=self._client,
                       index=self._index,
                       doc_type=self._doc_type,
                       mapping=self._mapping,
                       filter=self._filter,
                       groupby=self._groupby,
                       aggregation=_count_aggregator,
                       projection=self._projection,
                       sort=self._sort,
                       limit=self._limit,
                       compat=self._compat)
        return df

    def show(self, n=10000, truncate=15):
        """
        Prints the first ``n`` rows to the console.

        :param n:  Number of rows to show.
        :param truncate:  Number of words to be truncated for each column.

        >>> df.filter(df['age'] < 25).select('name').show(3)
        +------+
        | name |
        +------+
        | Alice|
        | Bob  |
        | Leo  |
        +------+
        """
        assert n > 0

        if self._aggregation:
            raise DataFrameException('show() is not allowed for aggregation. use collect() instead')

        query = self._execute()

        if self._projection:
            cols = [col.field_name() for col in self._projection]
        else:
            cols = self.columns

        if cols is None:
            raise _unbound_index_err

        sys.stdout.write(query.result_as_tabular(cols, n, truncate))
        sys.stdout.write('time: {0}ms\n'.format(query.millis_taken))

    def __repr__(self):
        if self.columns is None:
            return "DataFrame(Unbound)"
        return "DataFrame[%s]" % (", ".join("%s" % c for c in self.columns))

    def print_debug(self):
        """
        Post the query to the Elasticsearch Server and prints out the result it returned
        """
        if self._client is None:
            raise _unbound_index_err
        sys.stdout.write(json.dumps(self._client.post(data=self._build_query()), indent=4))

    def to_dict(self):
        """
        Converts the current :class:`DataFrame <DataFrame>` object to Elasticsearch search dictionary.

        :return: a dictionary which obeys the Elasticsearch RESTful protocol
        """
        return self._build_query()

    def print_schema(self):
        """
        Prints out the schema in the tree format.

        >>> df.print_schema()
        index_name
        |-- type_name
          |-- experience :  {'type': 'integer'}
          |-- id :  {'type': 'string'}
          |-- mobile :  {'index': 'not_analyzed', 'type': 'string'}
          |-- regions :  {'index': 'not_analyzed', 'type': 'string'}
        """
        if self._index is None:
            return

        sys.stdout.write('{0}\n'.format(self._index))
        index_name = list(self._mapping.keys())[0]
        if self._compat >= 7:
            json_obj = self._mapping[index_name]["mappings"]["properties"]
            sys.stdout.write(self.resolve_schema(json_obj))
        else:
            if self._doc_type is not None:
                json_obj = self._mapping[index_name]["mappings"][self._doc_type]["properties"]
                sys.stdout.write(self.resolve_schema(json_obj))
            else:
                raise DataFrameException('Please specify mapping for ES version under 7')

    def resolve_schema(self, json_prop, res_schema="", depth=1):
        for field in json_prop:
            if "properties" in json_prop[field]:
                res_schema += "{}|--{}:\n".format('  ' * depth, field)
                res_schema = self.resolve_schema(json_prop[field]["properties"],
                                                 res_schema, depth=depth+1)
            else:
                res_schema += "{}|--{}: {}\n".format('  ' * depth, field, json_prop[field])
        return res_schema

    def _build_query(self):
        query = dict()

        if self._limit:
            query['size'] = self._limit
        else:
            query['size'] = 20

        if self._groupby and not self._aggregation:
            query['aggregations'] = self._groupby
            query['size'] = 0

        if self._aggregation:
            if self._groupby is None:
                query['aggregations'] = self._aggregation
                query['size'] = 0

            else:
                agg = copy.deepcopy(self._groupby)
                # insert aggregator to the inner-most grouper
                inner_most = agg
                while True:
                    key = list(inner_most.keys())[0]
                    if 'aggregations' in inner_most[key]:
                        inner_most = inner_most[key]['aggregations']
                    else:
                        break
                key = list(inner_most.keys())[0]
                inner_most[key]['aggregations'] = self._aggregation
                query['aggregations'] = agg
                query['size'] = 0

        if self._filter:
            assert isinstance(self._filter, BooleanFilter)
            if self._compat >= 5:
                query['query'] = {'bool': {'filter': self._filter.build()}}
            else:
                query['query'] = {'filtered': {'filter': self._filter.build()}}

        if self._projection:
            query['_source'] = {"includes": [col.field_name() for col in self._projection], "excludes": []}

        if self._sort:
            query['sort'] = self._sort
        self._last_query = query
        return query

    def _get_cols(self, mapping):
        index = list(mapping.keys())[0]
        cols = self._get_mappings(mapping, index)

        if len(cols) == 0:
            raise DataFrameException('0 columns found in mapping')
        return cols

    @classmethod
    def resolve_mappings(cls, json_map):
        prop = []
        for field in json_map:
            nested_props = []
            if "properties" in json_map[field]:
                nested_props = cls.resolve_mappings(json_map[field]["properties"])
            if len(nested_props) == 0:
                prop.append(field)
            else:
                for nested_prop in nested_props:
                    prop.append("{}.{}".format(field, nested_prop))
        return prop

    def _get_mappings(self, json_map, index_name):
        if self._compat >= 7:
            return DataFrame.resolve_mappings(json_map[index_name]["mappings"]["properties"])
        else:
            if self._doc_type is not None:
                return DataFrame.resolve_mappings(json_map[index_name]["mappings"][self._doc_type]["properties"])
            else:
                raise DataFrameException('Please specify mapping for ES version under 7')
